@@ -14,6 +14,7 @@ use std::collections::{BTreeMap, BTreeSet};
 use std::ffi::{CStr, CString};
 use std::hash::{Hash, Hasher};
 use std::marker::PhantomData;
+use std::ops::{Bound, Range, RangeBounds};
 use std::os::raw::{c_char, c_void};
 use std::str::FromStr;
 use std::{fmt, ptr, slice, str};
@@ -86,6 +87,241 @@ extern "C" {
     ) -> *mut *const CodeModule;
 }
 
+/// A structure containing a set of disjoint ranges with attached contents.
+#[derive(Clone, Debug)]
+pub struct RangeMap<A, E> {
+    inner: BTreeMap<A, (A, E)>,
+}
+
+impl<A: Ord + Copy, E> RangeMap<A, E> {
+    /// Insert a range into the map.
+    ///
+    /// The range must be disjoint from all ranges that are already present.
+    /// Returns true if the insertion was successful.
+    pub fn insert(&mut self, range: Range<A>, contents: E) -> bool {
+        let low = self
+            .inner
+            .range(greater_than(range.start))
+            .next()
+            .map(|(key, _)| *key);
+        let high = self.inner.range(range.end..).next().map(|(key, _)| *key);
+
+        match (low, high) {
+            (None, _) => {
+                // all existing ranges end before the start of `range` => just insert range
+                self.inner.insert(range.end, (range.start, contents));
+                true
+            }
+
+            (Some(low_end), Some(high_end)) if low_end == high_end => {
+                let low_start = self.inner[&low_end].0;
+                if range.end <= low_start {
+                    self.inner.insert(range.end, (range.start, contents));
+                    true
+                } else {
+                    false
+                }
+            }
+
+            _ => false,
+        }
+    }
+
+    /// Retrieves the contents associated with the given address.
+    pub fn get_contents(&self, address: A) -> Option<&E> {
+        let (_, (lower_bound, entry)) = self.inner.range(greater_than(address)).next()?;
+
+        (*lower_bound <= address).then(move || entry)
+    }
+
+    /// Retrieves the contents associated with the given address, allowing mutation.
+    pub fn get_contents_mut(&mut self, address: A) -> Option<&mut E> {
+        let (_, (lower_bound, entry)) = self.inner.range_mut(greater_than(address)).next()?;
+
+        (*lower_bound <= address).then(move || entry)
+    }
+
+    /// Returns true if the given address is covered by some range in the map.
+    pub fn contains(&self, address: A) -> bool {
+        self.get_contents(address).is_some()
+    }
+}
+
+impl<A: Ord, E> Default for RangeMap<A, E> {
+    fn default() -> Self {
+        Self {
+            inner: BTreeMap::new(),
+        }
+    }
+}
+
+/// A structure representing a tree of disjoint ranges with associated contents.
+#[derive(Debug, Clone)]
+pub struct NestedRangeMap<A, E> {
+    lower_bound: Option<A>,
+    contents: Option<E>,
+    sub_ranges: BTreeMap<A, Box<NestedRangeMap<A, E>>>,
+}
+
+impl<A, E> NestedRangeMap<A, E>
+where
+    A: Ord + Copy + fmt::Debug,
+    E: Clone + fmt::Debug,
+{
+    /// Insert a range into the map.
+    ///
+    /// The insertion is valid if the new range does not
+    /// overlap nontrivially with any existing ranges
+    /// and is not equal to an existing range.
+    /// Returns true if the insertion was successful.
+    pub fn insert(&mut self, range: Range<A>, contents: E) -> bool {
+        let low = self
+            .sub_ranges
+            .range(greater_than(range.start))
+            .next()
+            .map(|(key, _)| *key);
+        let high = self
+            .sub_ranges
+            .range(range.end..)
+            .next()
+            .map(|(key, _)| *key);
+
+        match (low, high) {
+            (None, _) => {
+                // all existing ranges end before the start of `range` => just insert range
+                self.sub_ranges.insert(
+                    range.end,
+                    Box::new(NestedRangeMap::leaf(range.start, contents)),
+                );
+                true
+            }
+
+            (Some(low_end), None) => {
+                let low_start = self.sub_ranges[&low_end].lower_bound.unwrap();
+                if low_start < range.start {
+                    return false;
+                }
+                Self::insert_branch(&mut self.sub_ranges, range, contents, low_end..);
+                true
+            }
+            (Some(low_end), Some(high_end)) if low_end == high_end => {
+                let low_start = self.sub_ranges[&low_end].lower_bound.unwrap();
+                if range.start == low_start && range.end == low_end {
+                    return false;
+                }
+
+                if range.start >= low_start {
+                    self.sub_ranges
+                        .get_mut(&low_end)
+                        .unwrap()
+                        .insert(range, contents)
+                } else if range.end <= low_start {
+                    self.sub_ranges.insert(
+                        range.end,
+                        Box::new(NestedRangeMap::leaf(range.start, contents)),
+                    );
+                    true
+                } else if range.start < low_start && range.end == low_end {
+                    Self::insert_branch(&mut self.sub_ranges, range, contents, low_end..=low_end);
+                    true
+                } else {
+                    false
+                }
+            }
+
+            (Some(low_end), Some(high_end)) => {
+                let low_start = self.sub_ranges[&low_end].lower_bound.unwrap();
+                let high_start = self.sub_ranges[&high_end].lower_bound.unwrap();
+
+                if range.start > low_start {
+                    return false;
+                }
+
+                if range.end <= high_start {
+                    Self::insert_branch(&mut self.sub_ranges, range, contents, low_end..high_end);
+                    true
+                } else if range.end >= high_end {
+                    Self::insert_branch(&mut self.sub_ranges, range, contents, low_end..=high_end);
+                    true
+                } else {
+                    false
+                }
+            }
+        }
+    }
+
+    /// Retrieves the *most specific* contents associated with the given address, that is,
+    /// those associated with the smallest range that covers the address.
+    pub fn get_contents(&self, address: A) -> Option<&E> {
+        // Find the first subrange whose end is greater than `address. If `address` is in the map,
+        // it must be in this range.
+        let (_, sub_range) = self.sub_ranges.range(greater_than(address)).next()?;
+        let lb = sub_range.lower_bound.unwrap();
+        if lb <= address {
+            let contents = sub_range.contents.as_ref().unwrap();
+            let sub_contents = sub_range.get_contents(address);
+            Some(sub_contents.unwrap_or(contents))
+        } else {
+            None
+        }
+    }
+
+    /// Returns true if the given address is covered by some range in the map.
+    pub fn contains(&self, address: A) -> bool {
+        self.get_contents(address).is_some()
+    }
+
+    fn leaf(lower_bound: A, contents: E) -> Self {
+        Self {
+            lower_bound: Some(lower_bound),
+            contents: Some(contents),
+            sub_ranges: BTreeMap::default(),
+        }
+    }
+
+    /// Inserts the new entry(`range`, `contents`) into the map and moves
+    /// all entries with keys in `children` into the new entry.
+    fn insert_branch<R: RangeBounds<A>>(
+        sub_ranges: &mut BTreeMap<A, Box<NestedRangeMap<A, E>>>,
+        range: Range<A>,
+        contents: E,
+        children: R,
+    ) {
+        let mut sub_ranges_new = BTreeMap::new();
+        for (key, val) in sub_ranges.range(children) {
+            sub_ranges_new.insert(*key, val.clone());
+        }
+
+        for key in sub_ranges_new.keys() {
+            sub_ranges.remove(key);
+        }
+
+        sub_ranges.insert(
+            range.end,
+            Box::new(NestedRangeMap {
+                lower_bound: Some(range.start),
+                contents: Some(contents),
+                sub_ranges: sub_ranges_new,
+            }),
+        );
+    }
+}
+
+impl<A: Ord + Copy, E> Default for NestedRangeMap<A, E> {
+    fn default() -> Self {
+        Self {
+            lower_bound: None,
+            contents: None,
+            sub_ranges: BTreeMap::default(),
+        }
+    }
+}
+
+/// Creates the range (`lower_bound`, ∞).
+fn greater_than<A>(lower_bound: A) -> (Bound<A>, Bound<A>) {
+    (Bound::Excluded(lower_bound), Bound::Unbounded)
+}
+
 /// An error returned when parsing an invalid [`CodeModuleId`](struct.CodeModuleId.html).
 pub type ParseCodeModuleIdError = ParseDebugIdError;
 
@@ -621,7 +857,7 @@ impl SystemInfo {
             // Try to parse the Linux build string. Breakpad and Crashpad run
             // `uname -srvmo` to generate it. This roughtly resembles:
             // "Linux [version] [build...] [arch] Linux/GNU"
-            if let Some(captures) = LINUX_BUILD_RE.captures(&build) {
+            if let Some(captures) = LINUX_BUILD_RE.captures(build) {
                 let version = captures.get(1).unwrap(); // uname -r portion
                 let build = captures.get(2).unwrap(); // uname -v portion
                 return (version.as_str().into(), build.as_str().into());
@@ -969,3 +1205,146 @@ impl<'a> fmt::Debug for ProcessState<'a> {
             .finish()
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+    use proptest::prelude::*;
+
+    fn arb_disjoint_ranges(range: Range<u64>) -> impl Strategy<Value = Vec<Range<u64>>> {
+        fn go(range: Range<u64>, acc: &mut Vec<Range<u64>>) {
+            let mid = (range.end + range.start) / 2;
+            if range.start < mid && mid < range.end {
+                go(range.start..mid, acc);
+                go(mid..range.end, acc);
+            } else {
+                acc.push(range);
+            }
+        }
+
+        let mut ranges = Vec::new();
+        go(range, &mut ranges);
+
+        Just(ranges).prop_shuffle()
+    }
+
+    /// Creates a vector of nested subranges of `range` by recursively halving `range`
+    /// and shuffling the end result.
+    fn arb_nested_ranges(range: Range<u32>) -> impl Strategy<Value = Vec<Range<u32>>> {
+        fn go(range: Range<u32>, acc: &mut Vec<Range<u32>>) {
+            let mid = (range.end + range.start) / 2;
+            if range.start < mid && mid < range.end {
+                go(range.start..mid, acc);
+                go(mid..range.end, acc);
+            }
+
+            acc.push(range);
+        }
+
+        let mut ranges = Vec::new();
+        go(range, &mut ranges);
+
+        Just(ranges).prop_shuffle()
+    }
+
+    /// Checks that a `NestedRangeMap` is actually properly nested.
+    fn check<A: Ord + Copy + fmt::Debug, E>(map: NestedRangeMap<A, E>, range: Option<Range<A>>) {
+        for (upper, sub_map) in map.sub_ranges {
+            let lower = sub_map.lower_bound.unwrap();
+            if let Some(ref range) = range {
+                assert!(lower >= range.start, "{:?}, {:?}", range, lower..upper);
+                assert!(upper <= range.end, "{:?}, {:?}", range, lower..upper);
+            }
+            check(*sub_map, Some(lower..upper));
+        }
+    }
+
+    #[test]
+    fn nested_range_map_simple() {
+        let mut map = NestedRangeMap::default();
+
+        assert!(map.insert(0u8..10, "Outer"));
+        assert!(!map.insert(5..15, "Overlapping"));
+        assert!(map.insert(1..5, "Middle 1"));
+        assert!(map.insert(2..4, "Inner 1"));
+        assert!(map.insert(5..8, "Middle 2"));
+        assert!(!map.insert(3..8, "Overlapping"));
+        assert!(map.insert(6..8, "Inner 2"));
+        assert!(map.insert(0..9, "Middle 3"));
+
+        //  0    1    2    3    4    5    6    7    8    9    10
+        //          [Inner 1 ]          [Inner 2 ]
+        //     [   Middle 1       ][   Middle 2  ]
+        // [                  Middle 3                ]
+        // [                    Outer                      ]
+        assert_eq!(map.get_contents(0).unwrap(), &"Middle 3");
+        assert_eq!(map.get_contents(1).unwrap(), &"Middle 1");
+        assert_eq!(map.get_contents(2).unwrap(), &"Inner 1");
+        assert_eq!(map.get_contents(3).unwrap(), &"Inner 1");
+        assert_eq!(map.get_contents(4).unwrap(), &"Middle 1");
+        assert_eq!(map.get_contents(5).unwrap(), &"Middle 2");
+        assert_eq!(map.get_contents(6).unwrap(), &"Inner 2");
+        assert_eq!(map.get_contents(7).unwrap(), &"Inner 2");
+        assert_eq!(map.get_contents(8).unwrap(), &"Middle 3");
+        assert_eq!(map.get_contents(9).unwrap(), &"Outer");
+        assert_eq!(map.get_contents(10), None);
+    }
+
+    #[test]
+    fn nested_range_map_complex() {
+        let ranges = vec![
+            0..1,
+            1..2,
+            5..6,
+            3..4,
+            6..7,
+            3..5,
+            0..3,
+            5..7,
+            1..3,
+            4..5,
+            3..7,
+            0..7,
+            2..3,
+        ];
+
+        let mut map = NestedRangeMap::default();
+        for range in ranges {
+            assert!(map.insert(range, ()));
+        }
+
+        check(map, None);
+    }
+
+    proptest! {
+        #[test]
+        fn proptest_range_map(ranges in arb_disjoint_ranges(0..2^20)) {
+            let mut map = RangeMap::default();
+
+            for range in ranges.into_iter() {
+                assert!(
+                    map.insert(range, ()));
+            }
+
+            for i in 0..2^20 {
+                assert!(map.contains(i))
+            }
+        }
+
+        #[test]
+        fn proptest_nested_range_map(ranges in arb_nested_ranges(0..2^20)) {
+            let mut map = NestedRangeMap::default();
+
+            for range in ranges.into_iter() {
+                assert!(map.insert(range, ()));
+            }
+
+            for i in 0..2^20 {
+                assert!(map.contains(i))
+            }
+
+            check(map, None);
+
+        }
+    }
+}
